use std::collections::HashSet;

use anyhow::{anyhow, Error};
use bonsaidb::core::schema::{Collection, SerializedCollection};
use itertools::Itertools;
use serde::{Deserialize, Serialize};
use ubisync_lib::types::{Family, FamilyId, PeerId};

use crate::state::database::{as_key::AsKey, StateDB};

#[derive(Debug, Serialize, Deserialize, Collection, PartialEq, Clone)]
#[collection(name = "peer-families", views = [])]
pub(super) struct DbPeerFamily {
    #[natural_id]
    pub(super) id: AsKey<FamilyId>,
    pub(super) name: Option<String>,
    pub(super) members: HashSet<PeerId>,
}

impl From<DbPeerFamily> for Family {
    fn from(value: DbPeerFamily) -> Self {
        Family {
            id: (*value.id).clone(),
            name: value.name,
            members: value.members.iter().map(|p| p.clone()).collect_vec(),
        }
    }
}

impl StateDB {
    pub fn add_peer_family(
        &self,
        id: FamilyId,
        name: Option<String>,
        initial_members: Vec<PeerId>,
    ) -> anyhow::Result<()> {
        DbPeerFamily::push(
            DbPeerFamily {
                id: AsKey::new(id),
                name,
                members: HashSet::from_iter(initial_members),
            },
            &self.db,
        )
        .map(|_| ())
        .map_err(|e| anyhow!(e))
    }

    pub fn add_peer_to_family(&self, peer: PeerId, family: FamilyId) -> anyhow::Result<()> {
        DbPeerFamily::get(&AsKey::new(family), &self.db)
            .map_err(|e| anyhow!(e))?
            .ok_or(Error::msg("Could not find peer family"))?
            .modify(&self.db, |doc| {
                doc.contents.members.insert(peer.clone());
            })
            .map_err(|e| anyhow!(e))
    }

    pub fn set_peer_family_name(
        &self,
        family: FamilyId,
        name: Option<String>,
    ) -> anyhow::Result<()> {
        DbPeerFamily::get(&AsKey::new(family), &self.db)
            .map_err(|e| anyhow!(e))?
            .ok_or(Error::msg("Could not find peer family"))?
            .modify(&self.db, |doc| {
                doc.contents.name = name.clone();
            })
            .map_err(|e| anyhow!(e))
    }

    pub fn get_peer_family(&self, id: FamilyId) -> anyhow::Result<Option<Family>> {
        DbPeerFamily::get(&AsKey::new(id), &self.db)
            .map(|doc_opt| doc_opt.map(|doc| doc.contents.into()))
            .map_err(|e| anyhow!(e))
    }
}

#[cfg(test)]
mod tests {
    use ubisync_lib::types::{Family, FamilyId, PeerId};

    use crate::state::database::StateDB;

    #[test]
    fn add_get() {
        let db = StateDB::init(None);
        let family_id = FamilyId::new();
        let peer_id = PeerId::default();

        db.add_peer_family(
            family_id.clone(),
            Some("My family name".to_string()),
            vec![peer_id.clone()],
        )
        .unwrap();
        let retrieved_family = db.get_peer_family(family_id.clone()).unwrap();

        assert_eq!(
            retrieved_family,
            Some(Family::new(
                family_id,
                Some("My family name".to_string()),
                vec![peer_id]
            ))
        )
    }

    #[test]
    fn set_name() {
        let db = StateDB::init(None);
        let family_id = FamilyId::new();
        let peer_id = PeerId::default();

        db.add_peer_family(
            family_id.clone(),
            Some("My family name".to_string()),
            vec![peer_id.clone()],
        )
        .unwrap();

        assert_eq!(
            db.get_peer_family(family_id.clone()).unwrap(),
            Some(Family::new(
                family_id.clone(),
                Some("My family name".to_string()),
                vec![peer_id.clone()]
            ))
        );

        db.set_peer_family_name(family_id.clone(), Some("New family name".to_string()))
            .unwrap();

        assert_eq!(
            db.get_peer_family(family_id.clone()).unwrap(),
            Some(Family::new(
                family_id,
                Some("New family name".to_string()),
                vec![peer_id]
            ))
        );
    }
}
